/**
 * ModServ IRC Services
 * Version 2009
 * Copyright (C) 2007, 2008 Christian Jurk <commx@commx.ws>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 3 of the License,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <QtPlugin>
#include <MS_Dl.h>
#include <MS_Main.h>
#include <MS_Servers.h>
#include <MS_Channels.h>
#include "control.h"
#include <iostream>

#define MODULE_BASENAME  "control"
#define MODULE_NAME      "Control Service"
#define MODULE_VERSION   "1.0"
#define MODULE_SYS

/* global ModControl object */
ModControl *__ModControl;

ModControl::ModControl() {
  MConfigReader *cfg = globalMain->config;
  quint32 flags;

  m_modptr = 0;
  __ModControl = this;
  init();
  
  flags = MUser::parseModeflags(cfg->get("control.modes", "d"));

  controluser = new MVirtualUser(cfg->get("control.nick", "Control"));
  controluser->user->setIdent(cfg->get("control.ident", "control"));
  controluser->user->setHostname(cfg->get("control.host", "control.service"));
  controluser->setRealname(cfg->get("control.name", "ModServ Control Service"));
  controluser->user->setConnectedAt(QDateTime::currentDateTime());
  controluser->user->setModeflags(flags);
  controluser->setModule(m_modptr);
  
  if (!virtualServer || (virtualServer && virtualServer->state() < MServer::STATE_LINKED)) {
    connect(globalMain->parser, SIGNAL(onEndOfBurstAckMessage(const QStringList &)), this, SLOT(sendNickCreation(const QStringList &)));
  } else {
    controluser->create();
    controluser->joinChannel(cfg->get("control.channel", "#dev"), MChanMember::Flag_ChanOp);
    numericstr = MP10Parser::inttobase64(controluser->user->numeric(), 5);
    is_nickCreated = true;
  }
  
  connect(globalMain->parser, SIGNAL(onPrivmsgMessage(const QStringList &)), this, SLOT(handlePrivmsg(const QStringList &)));
}

ModControl::~ModControl() {
  disconnect(globalMain->parser, 0, this, 0);
}

void ModControl::init(void) {
  MModule *mptr;
  
  is_nickCreated = false;
  
  foreach(mptr, modList) {
    if (mptr->filePath().baseName() == QString("lib%1").arg(MODULE_BASENAME)) {
      mptr->setName(MODULE_NAME);
      mptr->setVersion(MODULE_VERSION);
      mptr->setBuilt(QString("%1 %2").arg(__DATE__).arg(__TIME__));
#ifdef MODULE_SYS
      mptr->setSysMod(true);
#endif
      m_modptr = mptr;
      break;
    }
  }
  
  registerCommand("get",
    "Displays a specified configuration item",
    "Usage: get var\nReplace \"var\" with the configuration identifier.",
    &c_get);

  registerCommand("help",
    "Displays help to a specific command",
    "Usage: help command\nWhere \"command\" has to replaced with the command you with help for.",
    &c_help);

  registerCommand("insmod",
    "Loads a shared module",
    "Usage: insmod path_to_module\nYou have to use relative or absolute filesystem paths to the module, including suffix.\nExample: insmod /home/users/commx/dev/modserv-2009.0/modules/control/libcontrol.so",
    &c_insmod);
    
  registerCommand("lsmod",
    "Lists all loaded modules",
    "Usage: lsmod",
    &c_lsmod);

  registerCommand("showcommands",
    "Displays a list, full with commands",
    "Usage: showcommands",
    &c_showcommands);
    
  registerCommand("rlmod",
    "Reloads a shared module",
    "Usage: rlmod path_to_module\nYou have to use relative or absolute filesystem paths to the module, including suffix.\nExample: rlmod modules/control/libcontrol.dynlib",
    &c_rlmod);

  registerCommand("rmmod",
    "Unloads a shared module",
    "Usage: rmmod path_to_module\nYou have to use relative or absolute filesystem paths to the module, including suffix.\nExample: rmmod modules/control/libcontrol.dll",
    &c_rmmod);
    
  registerCommand("serverlist",
    "Lists all connected servers on the network",
    "Usage: serverlist",
    &c_serverlist);

  registerCommand("set",
    "Sets or updates a specified configuration item",
    "Usage: set var value\nReplace \"var\" with the configuration identifier, followed by the value.",
    &c_set);
    
  registerCommand("showchan",
    "Displays informations about a channel",
    "Usage: showchan #channel\nReplace #channel with the channel you want informations from.",
    &c_showchan);

  registerCommand("stats",
    "Displays memory usage statistics",
    "Usage: stats",
    &c_stats);

  registerCommand("status",
    "Displays the service status",
    "Usage: status",
    &c_status);

  registerCommand("whois",
    "Displays informations about a user on the network",
    "Usage: whois nick\nReplace nick with the nick of the user you want informations from.",
    &c_whois);
}

void ModControl::sendNickCreation(const QStringList &strlArgs) {
  if (strlArgs.size() == 0)
    return;
  else if (!is_nickCreated) {
    MConfigReader *cfg = globalMain->config;
    controluser->create();
    controluser->joinChannel(cfg->get("control.channel", "#dev"), MChanMember::Flag_ChanOp);
    numericstr = MP10Parser::inttobase64(controluser->user->numeric(), 5);
    is_nickCreated = true;
  }
}

void ModControl::registerCommand(const QString &strCommand, const QString &strDesc, const QString &strHelp, cmdParam handler) {
  cmdHandler *hdl;
  
  hdl = new cmdHandler;
  hdl->command = strCommand;
  hdl->description = strDesc;
  hdl->helpmsg = strHelp;
  hdl->handler = handler;
  handlers.append(hdl);
}

void ModControl::deregisterCommand(const QString &strCommand) {
  foreach(cmdHandler *hdl, handlers) {
    if (hdl->command == strCommand) {
      handlers.removeOne(hdl);
      delete hdl;
    }
  }
}

void ModControl::handlePrivmsg(const QStringList &strlArgs) {
  MUser *user;
  
  if (strlArgs.at(2) != numericstr)
    return; /* message not originated to us */

  if (!(user = MUser::find(MP10Parser::base64toint(strlArgs.at(0)))))
    return; /* unable to locate that user */

  /* check permissions */
  if (!FLAG_SET(user->modeflags(), MUser::Flag_Operator)) {
    sendnotice(user, QString("You have not enough privileges to use this command."));
    return;
  }

  QStringList strlCmdParams;
  if (strlArgs.at(3).contains(" "))
    strlCmdParams = strlArgs.at(3).split(" ");
  else
    strlCmdParams << strlArgs.at(3);

  foreach(cmdHandler *hdl, handlers) {
    if (hdl->command == strlCmdParams.at(0).toLower()) {
      hdl->handler(user, strlCmdParams);
      return;
    }
  }

  sendnotice(user, QString("%1: unknown command. Try \"/msg %2 showcommands\" to get a command list overview.")
                      .arg(strlCmdParams.at(0))
                      .arg(controluser->user->nick()));
}

void ModControl::sendnotice(MUser *user, const QString &strMessage) {
  MTcpSocket *sock = globalMain->socket;
  sock->send(QString("%1 O %2 :%3").arg(numericstr).arg(MP10Parser::inttobase64(user->numeric(), 5)).arg(strMessage));
}

const QList<cmdHandler *> &ModControl::getHandlers(void) {
  return handlers;
}

Q_EXPORT_PLUGIN2(control, ModControl);
