/**
 * ModServ IRC Services
 * Version 2009
 * Copyright (C) 2007, 2008 Christian Jurk <commx@commx.ws>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 3 of the License,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <MS_Virtusers.h>
#include <MS_Main.h>
#include <MS_P10_Parser.h>
#include <MS_Channels.h>
#include <MS_Servers.h>

QList<MVirtualUser *> vuserList;

MVirtualUser::MVirtualUser(const QString &strNick) {
  user = new MUser;
  user->setNick(strNick);
  user->setNumeric(getFreeNumeric());
  user->setAccount(strNick);
  
  vuserList.append(this);
}

MVirtualUser::~MVirtualUser() {
  sendquit();
  userList.remove(user->numeric());
  nickList.remove(user->nick().toLower());
  delete user;
  vuserList.removeOne(this);
}

unsigned int MVirtualUser::getFreeNumeric(void) {
  MConfigReader *cfg = globalMain->config;
  unsigned int numindex, unum;
  QString base64, srvnum;

  srvnum = MP10Parser::inttobase64(cfg->get("vserver.numeric", "1").toUInt(), 2);
  
  for(numindex = 0; numindex < 262144; numindex++) {
    base64 = QString("%1%2").arg(srvnum).arg(MP10Parser::inttobase64(numindex, 3));
    unum = MP10Parser::base64toint(base64);
    
    if (!userList.contains(unum)) {
      bool found = false;
      foreach(MVirtualUser *pv, vuserList) {
        if (pv->user->numeric() == unum) {
          found = true;
          break;
        }
      }
      
      if (!found)
        return unum;
    }
  }
  
  return 0;
}

void MVirtualUser::setModule(MModule *module) {
  m_module = module;
}

void MVirtualUser::setRealname(const QString &strRealname) {
  m_realName = strRealname;
}

const QString &MVirtualUser::realname(void) {
  return m_realName;
}

MModule *MVirtualUser::module(void) {
  return m_module;
}

void MVirtualUser::create(void) {
  MConfigReader *cfg = globalMain->config;
  MTcpSocket *sock = globalMain->socket;
  QDateTime now = QDateTime::currentDateTime();
  QString protomsg;
  
  /* build the protocol message */
  protomsg = QString("%1 N %2 1 %3 %4 %5 +%6r %7 A%8 %9 :%10")
               .arg(MP10Parser::inttobase64(cfg->get("vserver.numeric", "1").toUInt(), 2))
               .arg(user->nick())
               .arg(now.toTime_t())
               .arg(user->ident())
               .arg(user->hostname())
               .arg(MUser::buildModeflagstr(user->modeflags()))
               .arg(user->account())
               .arg(MP10Parser::inttobase64(user->numeric(), 5))
               .arg(MP10Parser::inttobase64(user->numeric(), 5))
               .arg(realname());

  sock->send(protomsg);

  /* using this in the constructor caused QString::arg() to produce segfaults */
  userList.insert(user->numeric(), user);
  nickList.insert(user->nick().toLower(), user);
  user->setServer(virtualServer);
}

void MVirtualUser::joinChannel(const QString &strChannel, const quint32 &flags) {
  MTcpSocket *sock = globalMain->socket;
  MChannel *chan;
  QDateTime now = QDateTime::currentDateTime();
  QString protomsg;
  
  if (!(chan = MChannel::find(strChannel))) {
    chan = new MChannel(strChannel);
    chan->setCreationTS(now);
  }

  chan->addMember(user, flags);

  protomsg = QString("%1 J %2 :%3")
               .arg(user->numeric64())
               .arg(strChannel)
               .arg(now.toTime_t());
               
  sock->send(protomsg);
  
  if (flags != 0) {
    QStringList strlNums;
    QString mdList;
    
    if (FLAG_SET(flags, MChanMember::Flag_ChanOp)) {
      strlNums << MP10Parser::inttobase64(user->numeric(), 5);
      mdList.append("o");
    }
    if (FLAG_SET(flags, MChanMember::Flag_Voice)) {
      strlNums << MP10Parser::inttobase64(user->numeric(), 5);
      mdList.append("v");
    }
    
    if (strlNums.size() > 0) {
      protomsg = QString("%1 M %2 +%3 %4")
                   .arg(virtualServer->numeric64())
                   .arg(chan->name())
                   .arg(mdList)
                   .arg(strlNums.join(" "));
                   
      sock->send(protomsg);
    }
  }
}

void MVirtualUser::leaveChannel(const QString &strChannel) {
  MTcpSocket *sock = globalMain->socket;
  MChannel *chan;
  QDateTime now = QDateTime::currentDateTime();
  QString protomsg;

  if (!(chan = MChannel::find(strChannel)))
    return; /* channel does not exist */

  chan->removeMember(user);

  protomsg = QString("%1 L %2")
               .arg(user->numeric64())
               .arg(strChannel);

  sock->send(protomsg);
}

void MVirtualUser::sendquit(void) {
  MTcpSocket *sock = globalMain->socket;
  QString protomsg;
  
  protomsg = QString("%1 Q :Exiting").arg(MP10Parser::inttobase64(user->numeric(), 5));
  sock->send(protomsg);
}
